/* stph.c -- Sawmill-inspired Taskbar and Pager Helper
   
   Keep an internal representation of the state of tasks as required by taskbars/pagers; report
   changes in state; generally take care of all the low-level X stuff required to do so, in order to
   make it simple to concentrate on the UI side of things when writing taskbars/pagers...

   Copyright (C) 2000 Rob Hodges <s323140@student.uq.edu.au>

   This file is part of stph.

   stph is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   stph is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with stph; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */


/* Some parts of stph are derived from the gwmh library used by Gnome pagers,
   which was written by Tim Janik */

/* ---------- Includes: ---------- */
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include "stph.h"

/* ---------- Defines; Macros: ---------- */
#ifndef min
#define min(a,b)  ((a) < (b) ? (a) : (b))
#endif

#define MAXTASKS 200 /* FIXME: stupid, arbitrary limit; should use a
                        dynamically allocated linked list instead */

#define RUN_CALLBACK(cb_type,cb_arg) callbacks[cb_type].func(cb_arg, callbacks[cb_type].data)

/* ---------- Structs: ---------- */

typedef struct
{
        void (*func)();
        void *data;
} func_and_data;

typedef struct
{
        int flag;
        XErrorEvent err;
} xerror_data;


/* ---------- Globals: ---------- */
static char *program_name = "stph"; /*Sawmill-inspired taskbar and pager helper*/

static Display *dpy;
static int screen;

static XContext tasks_context;  /* this is for efficiently looking up tasks from xwins */
static struct StphTask *tasks[MAXTASKS]; /* this is for iterating over all tasks */
static struct StphRoot rootprops;
static int numtasks;

static xerror_data last_error;

static func_and_data callbacks[STPH_MAX_ROOT_CALLBACK+1];

static Atom STPH_WIN_SUPPORTING_WM_CHECK;
static Atom STPH_WIN_PROTOCOLS;
static Atom STPH_WIN_LAYER;
static Atom STPH_WIN_STATE;
static Atom STPH_WIN_HINTS;
static Atom STPH_WIN_APP_STATE;
static Atom STPH_WIN_EXPANDED_SIZE;
static Atom STPH_WIN_ICONS;
static Atom STPH_WIN_WORKSPACE;
static Atom STPH_WIN_WORKSPACE_COUNT;
static Atom STPH_WIN_WORKSPACE_NAMES;
static Atom STPH_WIN_CLIENT_LIST;
static Atom STPH_WIN_AREA;
static Atom STPH_WIN_AREA_COUNT;
static Atom STPH_WIN_UNIFIED_AREA;
static Atom STPH_KWM_WIN_ICON;
static Atom XA_WM_STATE;
static Atom XA_WM_PROTOCOLS;
static Atom XA_WM_TAKE_FOCUS;
static Atom XA_WM_DELETE_WINDOW;


/* ---------- Prototypes: ---------- */

/* Support functions: */
static int Get_Window_Property(Window target_win, Atom atom, long *length, Atom *type, int *size,
                        char **data);
static void Setup_Atoms(void);
static Window Get_Task_Frame(Window xwin); /* returns 0 if can't find frame */
static int wm_protocol_supported_p(Window xwin, Atom a);
static int send_client_message_32 (Window recipient,
                                   Window event_window,
                                   Atom message_type,
                                   long event_mask,
                                   int n_longs,
                                   ...);

/* Task alteration handlers: */
static int Add_Task(Window xwin);
static int Remove_Task(Window xwin);
static int Refresh_Tasks(void);
static void Update_Task_Name(struct StphTask *t, int inhibit_callbacks);
static void Update_Task_Icon_Name(struct StphTask *t, int inhibit_callbacks);
static void Update_Task_Class(struct StphTask *t, int inhibit_callbacks);
static void Update_Task_State(struct StphTask *t, int inhibit_callbacks);
static void Update_Task_Hints(struct StphTask *t, int inhibit_callbacks);
static void Update_Task_WM_Hints(struct StphTask *t, int inhibit_callbacks);
static void Update_Task_KWM_Icon(struct StphTask *t, int inhibit_callbacks);
static void Update_Task_Focused(struct StphTask *t, int inhibit_callbacks);
static void Update_Task_Iconified(struct StphTask *t, int inhibit_callbacks);
static void Update_Task_Frame(struct StphTask *t, int inhibit_callbacks);
static void Update_Task_Win(struct StphTask *t, int inhibit_callbacks);
static void Update_Task_Layer(struct StphTask *t, int inhibit_callbacks);
static void Update_Task_Desktop(struct StphTask *t, int inhibit_callbacks);
static void Update_Task_Area(struct StphTask *t, int inhibit_callbacks);

/* Root window alteration handlers: */
static void Update_Compliant_WM(void);
static void Update_Supported_Protocols(void);
static void Update_Win_Area(int inhibit_callbacks);
static void Update_Win_Area_Count(int inhibit_callbacks);
static void Update_Win_Workspace(int inhibit_callbacks);
static void Update_Win_Workspace_Count(int inhibit_callbacks);

/* Event handler: */
static int Process_Event(XEvent *ev);

/* Misc. private functions: */
static int check_errors(Display *disp, XErrorEvent *xerr);
static void print_last_error(void);


static int Get_Window_Property(Window target_win, Atom atom, long *length, Atom *type, int *size,
                        char **data)
{
  Atom actual_type;
  int actual_format;
  unsigned long nitems;
  unsigned long nbytes;
  unsigned long bytes_after;
  unsigned char *prop;
  int status;

  int max_len = 10000;

  status = XGetWindowProperty(dpy, target_win, atom, 0, (max_len+3)/4,
                              False, AnyPropertyType, &actual_type,
                              &actual_format, &nitems, &bytes_after,
                              &prop);

  
  if (last_error.flag) {
          last_error.flag = 0;
          return 1;
  }

  if (status!=Success) /* some other problem... */
          return 1;

  if (actual_format == 32)
          nbytes = sizeof(long);
  else if (actual_format == 16)
          nbytes = sizeof(short);
  else
          nbytes = 1;
  *length = min(nitems * nbytes, max_len);
  *type = actual_type;
  *size = actual_format;
  *data = prop;

  if (prop == NULL)
          return 1;

  return 0;
}

static void Setup_Atoms(void)
{
  /* FIXME: Should be using XInternAtoms to reduce traffic */
  STPH_WIN_SUPPORTING_WM_CHECK = XInternAtom(dpy, "_WIN_SUPPORTING_WM_CHECK", False);
  STPH_WIN_PROTOCOLS = XInternAtom(dpy, "_WIN_PROTOCOLS", False);
  STPH_WIN_LAYER = XInternAtom(dpy, "_WIN_LAYER", False);
  STPH_WIN_STATE = XInternAtom(dpy, "_WIN_STATE", False);
  STPH_WIN_HINTS = XInternAtom(dpy, "_WIN_HINTS", False);
  STPH_WIN_APP_STATE = XInternAtom(dpy, "_WIN_APP_STATE", False);
  STPH_WIN_EXPANDED_SIZE = XInternAtom(dpy, "_WIN_EXPANDED_SIZE", False);
  STPH_WIN_ICONS = XInternAtom(dpy, "_WIN_ICONS", False);
  STPH_WIN_WORKSPACE = XInternAtom(dpy, "_WIN_WORKSPACE", False);
  STPH_WIN_WORKSPACE_COUNT = XInternAtom(dpy, "_WIN_WORKSPACE_COUNT", False);
  STPH_WIN_WORKSPACE_NAMES = XInternAtom(dpy, "_WIN_WORKSPACE_NAMES", False);
  STPH_WIN_CLIENT_LIST = XInternAtom(dpy, "_WIN_CLIENT_LIST", False);
  STPH_WIN_AREA = XInternAtom(dpy, "_WIN_AREA", False);
  STPH_WIN_AREA_COUNT = XInternAtom(dpy, "_WIN_AREA_COUNT", False);
  STPH_WIN_UNIFIED_AREA = XInternAtom(dpy, "_WIN_UNIFIED_AREA", False);
  STPH_KWM_WIN_ICON = XInternAtom(dpy, "KWM_WIN_ICON", False);
  XA_WM_STATE = XInternAtom(dpy, "WM_STATE", False);
  XA_WM_PROTOCOLS = XInternAtom(dpy, "WM_PROTOCOLS", False);
  XA_WM_TAKE_FOCUS = XInternAtom(dpy, "WM_TAKE_FOCUS", False);
  XA_WM_DELETE_WINDOW = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
}

struct StphTask *stph_task_from_xwin(Window xwin)
{
  XPointer t = NULL;
  
  if (XFindContext(dpy, xwin, tasks_context, &t))
          return NULL;
  return (struct StphTask *) t;
}

struct StphRoot *stph_get_root_props(void)
{
  return &rootprops;
}


static Window Get_Task_Frame(Window xwin)
{
  Window *children = NULL, xframe = 0;
  Window xparent = xwin, xroot = xwin;
  unsigned int size = 0;

  while (XQueryTree (dpy, xwin, &xroot, &xparent, &children, &size)) {
          if (children) {
                  XFree (children);
                  children = NULL;
          }
          
          xframe = xwin;
          xwin = xparent;
          if (xparent == xroot)
                  break;
  }
  if (last_error.flag)
          last_error.flag = 0; /* Ignore BadWindow here */
  
  return xframe;
}

static void Update_Task_Name(struct StphTask *t, int inhibit_callbacks)
{
  char *val;
  long length;
  Atom type;
  int size;
  char *oldname;

  if (t == NULL)
          return;

  oldname = (char *) malloc(strlen(t->name)+1);
  strcpy(oldname, t->name);
  
  free(t->name);
  if (Get_Window_Property(t->xwin, XA_WM_NAME, &length, &type, &size, &val)) {
          t->name = (char *) malloc(4);
          strcpy(t->name, "???");
  } else {
          t->name = (char *) malloc(strlen(val)+1);
          strcpy(t->name, val);
          XFree(val);
  }

  if (callbacks[STPH_TASK_NAME_CB].func && !inhibit_callbacks && strcmp(t->name, oldname))
          RUN_CALLBACK(STPH_TASK_NAME_CB, t);
  free(oldname);
}

static void Update_Task_Icon_Name(struct StphTask *t, int inhibit_callbacks)
{
  char *val;
  long length;
  Atom type;
  int size;
  char *oldname;

  if (t == NULL)
          return;

  oldname = (char *) malloc(strlen(t->icon_name)+1);
  strcpy(oldname, t->icon_name);
  
  free(t->icon_name);
  if (Get_Window_Property(t->xwin, XA_WM_ICON_NAME, &length, &type, &size, &val)) {
          t->icon_name = (char *) malloc(4);
          strcpy(t->icon_name, "???");
  } else {
          t->icon_name = (char *) malloc(strlen(val)+1);
          strcpy(t->icon_name, val);
          XFree(val);
  }

  if (callbacks[STPH_TASK_ICON_NAME_CB].func && !inhibit_callbacks && strcmp(t->icon_name, oldname))
          RUN_CALLBACK(STPH_TASK_ICON_NAME_CB, t);
  free(oldname);
}


static void Update_Task_Class(struct StphTask *t, int inhibit_callbacks)
{
  char *val;
  long length;
  Atom type;
  int size;
  char *oldclass;
  char *oldinst;
  long instlen;

  if (t == NULL)
          return;

  oldclass = (char *) malloc(strlen(t->class)+1);
  strcpy(oldclass, t->class);
  
  oldinst = (char *) malloc(strlen(t->instance)+1);
  strcpy(oldinst, t->instance);
  
  free(t->class);
  free(t->instance);

  if (Get_Window_Property(t->xwin, XA_WM_CLASS, &length, &type, &size, &val)) {
          t->class = (char *) malloc(4);
          strcpy(t->class, "???");
          t->instance = (char *) malloc(4);
          strcpy(t->instance, "???");
  } else {
          instlen = strlen(val);
          t->instance = (char *) malloc(instlen+1);
          strcpy(t->instance, val);
          t->class = (char *) malloc(strlen(val+instlen+1)+1);
          strcpy(t->class, val+instlen+1);
          XFree(val);
  }

  if (callbacks[STPH_TASK_CLASS_CB].func && !inhibit_callbacks
      && (strcmp(t->class, oldclass) || strcmp(t->instance, oldinst)))
          RUN_CALLBACK(STPH_TASK_CLASS_CB, t);
  free(oldclass);
  free(oldinst);
}

static void Update_Task_State(struct StphTask *t, int inhibit_callbacks)
{
  char *val;
  long length;
  Atom type;
  int size;
  StphState oldstate;

  if (t == NULL)
          return;

  oldstate = t->state;
  
  if (Get_Window_Property(t->xwin, STPH_WIN_STATE, &length, &type, &size, &val)) {
          t->state = 0;
  } else {
          t->state = (long) * (long *) val;
          XFree(val);
  }

  if (callbacks[STPH_TASK_STATE_CB].func && !inhibit_callbacks && oldstate != t->state)
          RUN_CALLBACK(STPH_TASK_STATE_CB,t);
}

static void Update_Task_Hints(struct StphTask *t, int inhibit_callbacks)
{
  char *val;
  long length;
  Atom type;
  int size;
  StphHints oldhints;

  if (t == NULL)
          return;

  oldhints = t->hints;
  
  if (Get_Window_Property(t->xwin, STPH_WIN_HINTS, &length, &type, &size, &val)) {
          t->hints = 0;
  } else {
          t->hints = (long) * (long *) val;
          XFree(val);
  }

  if (callbacks[STPH_TASK_HINTS_CB].func && !inhibit_callbacks && oldhints != t->hints)
          RUN_CALLBACK(STPH_TASK_HINTS_CB,t);
}

static void Update_Task_WM_Hints(struct StphTask *t, int inhibit_callbacks)
{
  char *val;
  long length;
  Atom type;
  int size;
  XWMHints *hints;
  Pixmap old_pixmap, old_mask;
  Window old_window;

  if (t == NULL)
          return;

  old_pixmap = t->icon_pixmap;
  old_mask = t->icon_mask;
  old_window = t->icon_window;
  
  if (Get_Window_Property(t->xwin, XA_WM_HINTS, &length, &type, &size, &val)) {
          t->icon_pixmap = 0;
          t->icon_mask = 0;
          t->icon_window = 0;
  } else {
          hints = (XWMHints *) val;
          if (hints->flags & IconPixmapHint)
                  t->icon_pixmap = hints->icon_pixmap;
          else
                  t->icon_pixmap = 0;
          if (hints->flags & IconMaskHint)
                  t->icon_mask = hints->icon_mask;
          else
                  t->icon_mask = 0;
          if (hints->flags & IconWindowHint)
                  t->icon_window = hints->icon_window;
          else
                  t->icon_window = 0;
          XFree(val);
  }

  if (callbacks[STPH_TASK_ICON_PIXMAP_CB].func && !inhibit_callbacks &&
      (old_pixmap != t->icon_pixmap || old_mask != t->icon_mask))
          RUN_CALLBACK(STPH_TASK_ICON_PIXMAP_CB, t);
  
  if (callbacks[STPH_TASK_ICON_WINDOW_CB].func && !inhibit_callbacks &&
      old_window != t->icon_window)
          RUN_CALLBACK(STPH_TASK_ICON_WINDOW_CB, t);
}

static void Update_Task_KWM_Icon(struct StphTask *t, int inhibit_callbacks)
{
  char *val;
  long length;
  Atom type;
  int size;
  Pixmap old_icon, old_mask;
  Pixmap *prop;

  if (t == NULL)
          return;

  old_icon = t->kwm_icon_pixmap;
  old_mask = t->kwm_icon_mask;

  if (Get_Window_Property(t->xwin, STPH_KWM_WIN_ICON, &length, &type, &size, &val)) {
          t->kwm_icon_pixmap = 0;
          t->kwm_icon_mask = 0;
  } else {
          prop = (Pixmap *) val;
          t->kwm_icon_pixmap = prop[0];
          t->kwm_icon_mask = prop[1];
          XFree(val);
  }

  if (callbacks[STPH_TASK_KWM_ICON_CB].func && !inhibit_callbacks &&
      (old_icon != t->kwm_icon_pixmap || old_mask != t->kwm_icon_mask))
          RUN_CALLBACK(STPH_TASK_KWM_ICON_CB, t);
}


static void Update_Task_Focused(struct StphTask *t, int inhibit_callbacks)
{
  Window focus;
  int revert_to;
  int oldfocused;

  if (t == NULL)
          return;

  oldfocused = t->focused;

  XGetInputFocus(dpy, &focus, &revert_to);
  
  if (last_error.flag) {
          last_error.flag = 0;
          return; /* ignore BadWindow etc. here */
  }
  
  if (focus == t->xwin)
          t->focused = True;
  else
          t->focused = False;

  if (callbacks[STPH_TASK_FOCUS_CB].func && !inhibit_callbacks && oldfocused != t->focused)
          RUN_CALLBACK(STPH_TASK_FOCUS_CB,t);
}

static void Update_Task_Iconified(struct StphTask *t, int inhibit_callbacks)
{
  char *val;
  long length;
  Atom type;
  int size;

  int old;

  if (t == NULL)
          return;

  old = t->iconified;
  
  if (Get_Window_Property(t->xwin, XA_WM_STATE, &length, &type, &size, &val)) {
          t->iconified = False;
  } else {
          t->iconified = ( ((long *) val)[0] == 3 );
          XFree(val);
  }

  if (callbacks[STPH_TASK_ICONIFY_CB].func && !inhibit_callbacks && old != t->iconified)
          RUN_CALLBACK(STPH_TASK_ICONIFY_CB,t);
}

static void Update_Task_Frame(struct StphTask *t, int inhibit_callbacks)
{
  XWindowAttributes fattribs = { 0, };
  Window old_frame;
  int old_x, old_y, old_w, old_h;
  
  if (t == NULL)
          return;

  old_frame = t->xframe;
  old_x = t->frame_x;
  old_y = t->frame_y;
  old_w = t->frame_width;
  old_h = t->frame_height;
  
  t->xframe = Get_Task_Frame(t->xwin);
  if (!(t->xframe)) {
          t->xframe = old_frame;
          return;
  }
  
  XGetWindowAttributes(dpy, t->xframe, &fattribs);
  /* We should only get BadWindow here, if the frame ceased to exist since the immediately preceding */
  /* call to Get_Task_Frame().  In which case, it probably doesn't matter and we'll just return for */
  /* now, coz we'll hear about that soon enough. */
  if (last_error.flag) {
          last_error.flag = 0;
          return;
  }
  t->frame_x = fattribs.x;
  t->frame_y = fattribs.y;
  t->frame_width = fattribs.width;
  t->frame_height = fattribs.height;

  if (callbacks[STPH_TASK_FRAME_CB].func && !inhibit_callbacks && old_frame != t->xframe)
          RUN_CALLBACK(STPH_TASK_FRAME_CB,t);
  if (callbacks[STPH_TASK_FRAME_POS_CB].func && !inhibit_callbacks && (old_x != t->frame_x || old_y != t->frame_y))
          RUN_CALLBACK(STPH_TASK_FRAME_POS_CB,t);
  if (callbacks[STPH_TASK_FRAME_DIMS_CB].func && !inhibit_callbacks
      && (old_w != t->frame_width || old_h != t->frame_height))
          RUN_CALLBACK(STPH_TASK_FRAME_DIMS_CB,t);
}

static void Update_Task_Win(struct StphTask *t, int inhibit_callbacks)
{
  XWindowAttributes wattribs = { 0, };
  int old_x, old_y, old_w, old_h;

  if (t == NULL)
          return;

  old_x = t->win_x;
  old_y = t->win_y;
  old_w = t->win_width;
  old_h = t->win_height;

  XGetWindowAttributes(dpy, t->xwin, &wattribs);
  /* If we get BadWindow here, the window must have just disappeared.  No point continuing with */
  /* this; we'll hear about that soon enough. */
  if (last_error.flag) {
          last_error.flag = 0;
          return;
  }
  t->win_x = wattribs.x; /* wrt parent (ie, frame) */
  t->win_y = wattribs.y; /* wrt parent (ie, frame) */
  t->win_width = wattribs.width;
  t->win_height = wattribs.height;

  if (callbacks[STPH_TASK_WIN_POS_CB].func && !inhibit_callbacks && (old_x != t->win_x || old_y != t->win_y))
          RUN_CALLBACK(STPH_TASK_WIN_POS_CB,t);
  if (callbacks[STPH_TASK_WIN_DIMS_CB].func && !inhibit_callbacks && (old_w != t->win_width || old_h != t->win_height))
          RUN_CALLBACK(STPH_TASK_WIN_DIMS_CB,t);
}

static void Update_Task_Layer(struct StphTask *t, int inhibit_callbacks)
{
  char *val;
  long length;
  Atom type;
  int size;

  StphLayer old;

  if (t == NULL)
          return;

  old = t->layer;
  
  if (Get_Window_Property(t->xwin, STPH_WIN_LAYER, &length, &type, &size, &val)) {
          t->layer = 4;
  } else {
          t->layer = (long) * (long *) val;
          XFree(val);
  }

  if (callbacks[STPH_TASK_LAYER_CB].func && !inhibit_callbacks && old != t->layer)
          RUN_CALLBACK(STPH_TASK_LAYER_CB,t);
}

static void Update_Task_Desktop(struct StphTask *t, int inhibit_callbacks)
{
  char *val;
  long length;
  Atom type;
  int size;

  int old;

  if (t == NULL)
          return;

  old = t->desktop;
  
  if (Get_Window_Property(t->xwin, STPH_WIN_WORKSPACE, &length, &type, &size, &val)) {
          t->desktop = 0;
  } else {
          t->desktop = (long) * (long *) val;
          XFree(val);
  }

  if (callbacks[STPH_TASK_DESKTOP_CB].func && !inhibit_callbacks && old != t->desktop)
          RUN_CALLBACK(STPH_TASK_DESKTOP_CB,t);
}

static void Update_Task_Area(struct StphTask *t, int inhibit_callbacks)
{
  char *val;
  long length;
  Atom type;
  int size;

  int old_h, old_v;

  if (t == NULL)
          return;

  old_h = t->harea;
  old_v = t->varea;
  
  if (Get_Window_Property(t->xwin, STPH_WIN_AREA, &length, &type, &size, &val)) {
          t->harea = 0;
          t->varea = 0;
  } else {
          t->harea = ((long *) val)[0];
          t->varea = ((long *) val)[1];
          XFree(val);
  }

  if (callbacks[STPH_TASK_AREA_CB].func && !inhibit_callbacks && (old_h != t->harea || old_v != t->varea))
          RUN_CALLBACK(STPH_TASK_AREA_CB,t);
}

static int Add_Task(Window xwin)
{
  struct StphTask *t;

  /* check if we already know about this task */
  if (stph_task_from_xwin(xwin) != NULL)
          return 0;
  
  /* check if user has insane number of windows open */
  if (numtasks == MAXTASKS) {
          fprintf(stderr, "%s: MAXTASKS reached!\n", program_name);
          return 1;
  }

  /* allocate memory for task and task.name, task.icon_name, task.class, task.instance */
  t = (struct StphTask *) malloc(sizeof(struct StphTask));

  t->name = (char *) malloc(4);
  strcpy(t->name, "???");

  t->icon_name = (char *) malloc(4);
  strcpy(t->icon_name, "???");

  t->class = (char *) malloc(4);
  strcpy(t->class, "???");

  t->instance = (char *) malloc(4);
  strcpy(t->instance, "???");
  
  /* setup all of the task's properties */
  t->xwin = xwin;

  Update_Task_Name(t, True);
  Update_Task_Icon_Name(t, True);
  Update_Task_Class(t, True);
  Update_Task_State(t, True);
  Update_Task_Hints(t, True);
  Update_Task_WM_Hints(t, True);
  Update_Task_KWM_Icon(t, True);
  Update_Task_Focused(t, True);
  Update_Task_Iconified(t, True);
  Update_Task_Frame(t, True);
  Update_Task_Win(t, True);
  Update_Task_Layer(t, True);
  Update_Task_Desktop(t, True);
  Update_Task_Area(t, True);

  /* Tell X what events we want for this window */
  XSelectInput(dpy, xwin, PropertyChangeMask | FocusChangeMask | StructureNotifyMask);
  XSelectInput(dpy, t->xframe, StructureNotifyMask);
  /* XSelectInput() can generate BadWindow errors; in that case the window must have already
     disappeared.  So free the memory and don't do anything more. */
  if (last_error.flag) {
          free(t->name);
          free(t->icon_name);
          free(t->class);
          free(t->instance);
          free(t);
          last_error.flag = 0;
          return 1;
  }

  /* hash t to its xwin for quick retrieval later; if this doesn't succeed we're better off
     forgetting about it (but it should succeed) */
  if (XSaveContext(dpy, t->xwin, tasks_context, (XPointer) t)) {
          free(t->name);
          free(t->icon_name);
          free(t->class);
          free(t->instance);
          free(t);
          return 1;
  }

  tasks[numtasks++] = t;

  if (callbacks[STPH_ADD_TASK_CB].func)
          RUN_CALLBACK(STPH_ADD_TASK_CB,t);
  
  return 0;
}

static int Remove_Task(Window xwin)
{
  int i;

  if (callbacks[STPH_REMOVE_TASK_CB].func)
          RUN_CALLBACK(STPH_REMOVE_TASK_CB,stph_task_from_xwin(xwin));

  XDeleteContext(dpy, xwin, tasks_context); /* there _shouln't_ be an error here,
                                               so we won't check the return code.
                                               FIXME: really we should put an assert. */
  for (i=0; i<numtasks; i++) {
          if (tasks[i]->xwin == xwin) {
                  free(tasks[i]->name);
                  free(tasks[i]->icon_name);
                  free(tasks[i]->class);
                  free(tasks[i]->instance);
                  free(tasks[i]);
                  if (i != --numtasks) {
                          tasks[i] = tasks[numtasks];
                  }
                  return 0;
          }
  }
  return 1;
}


/* FIXME: eventually we should make this able to deal with _WIN_CLIENT_LIST not being available, so
   that with reduced functionality we can support even non-gnome-compliant wm's... */
static int Refresh_Tasks(void)
{
  char *val;
  long length;
  Atom type;
  int size;
  long* clients;
  int i, j, nclients;
  

  if (Get_Window_Property(RootWindow(dpy, screen), STPH_WIN_CLIENT_LIST, &length, &type, &size, &val))
          return 0; /* assume that we've just started up and no tasks exist yet */

  clients = (long *) val;
  nclients = length / sizeof(long);

  /* add any tasks we didn't know about: */
  for (i=0; i<nclients; i++)
          Add_Task(clients[i]);

  /* remove any tasks not in clients: */
  for (i=0; i<numtasks; i++) {
          for (j=0; j<nclients && clients[j] != tasks[i]->xwin; j++)
                  ;
          if (j == nclients)
                  Remove_Task(tasks[i]->xwin);
  }
  
  XFree(val);

  /* we miss FocusIn events on newly-created windows, so... */
  for (i=0; i<numtasks; i++)
          Update_Task_Focused(tasks[i], False);
  
  return 0;
}

static void Update_Compliant_WM(void)
{
  char *val;
  long length;
  Atom type;
  int size;
  Window supp;
  
  /* _WIN_SUPPORTING_WM_CHECK must be set on the root window... */
  if (Get_Window_Property(RootWindow(dpy,screen),
                          STPH_WIN_SUPPORTING_WM_CHECK, &length,
                          &type, &size, &val)) {
          rootprops.compliant_wm = 0;
  } else {
          /* ... and furthermore, it must be set to an xwin, which
             must itself have _WIN_SUPPORTING_WM_CHECK set on it too
             (if this window goes away, the WM has died, so even if
             the property is still set on the root window, we don't
             have a compliant WM anymore. */
          supp = (long) * (long *) val;
          XFree(val);
          if (Get_Window_Property(supp, STPH_WIN_SUPPORTING_WM_CHECK,
                                  &length, &type, &size, &val)) {
                  rootprops.compliant_wm = 0;
          } else {
                  rootprops.compliant_wm = 1;
                  XFree(val);
          }
  }
}


static void Update_Supported_Protocols(void)
{
  char *val;
  long length;
  Atom type;
  int size;
  Atom *prots;
  int nprots, i;

  
  rootprops.supported_protocols = 0;
  
  if (!(Get_Window_Property(RootWindow(dpy,screen), STPH_WIN_PROTOCOLS, &length, &type,
                            &size, &val))) {
          prots = (Atom *) val;
          nprots = length / sizeof(Atom);
          for (i=0; i<nprots; i++) {
                  if (prots[i] == STPH_WIN_LAYER)
                          rootprops.supported_protocols |= STPH_WIN_LAYER_PCOL;
                  else if (prots[i] == STPH_WIN_STATE)
                          rootprops.supported_protocols |= STPH_WIN_STATE_PCOL;
                  else if (prots[i] == STPH_WIN_HINTS)
                          rootprops.supported_protocols |= STPH_WIN_HINTS_PCOL;
                  else if (prots[i] == STPH_WIN_APP_STATE)
                          rootprops.supported_protocols |= STPH_WIN_APP_STATE_PCOL;
                  else if (prots[i] == STPH_WIN_EXPANDED_SIZE)
                          rootprops.supported_protocols |= STPH_WIN_EXPANDED_SIZE_PCOL;
                  else if (prots[i] == STPH_WIN_ICONS)
                          rootprops.supported_protocols |= STPH_WIN_ICONS_PCOL;
                  else if (prots[i] == STPH_WIN_WORKSPACE)
                          rootprops.supported_protocols |= STPH_WIN_WORKSPACE_PCOL;
                  else if (prots[i] == STPH_WIN_WORKSPACE_COUNT)
                          rootprops.supported_protocols |= STPH_WIN_WORKSPACE_COUNT_PCOL;
                  else if (prots[i] == STPH_WIN_WORKSPACE_NAMES)
                          rootprops.supported_protocols |= STPH_WIN_WORKSPACE_NAMES_PCOL;
                  else if (prots[i] == STPH_WIN_CLIENT_LIST)
                          rootprops.supported_protocols |= STPH_WIN_CLIENT_LIST_PCOL;
          }
          XFree(val);
  }
}

static void Update_Win_Area(int inhibit_callbacks)
{
  char *val;
  long length;
  Atom type;
  int size;

  int old_h, old_v;

  old_h = rootprops.harea;
  old_v = rootprops.varea;
  
  if (Get_Window_Property(RootWindow(dpy,screen), STPH_WIN_AREA, &length, &type, &size, &val)) {
          rootprops.harea = 0;
          rootprops.varea = 0;
  } else {
          rootprops.harea = ((long *) val)[0];
          rootprops.varea = ((long *) val)[1];
          XFree(val);
  }

  if (callbacks[STPH_AREA_CB].func && !inhibit_callbacks
      && (old_h != rootprops.harea || old_v != rootprops.varea))
          RUN_CALLBACK(STPH_AREA_CB,&rootprops);
}

static void Update_Win_Area_Count(int inhibit_callbacks)
{
  char *val;
  long length;
  Atom type;
  int size;

  int old_h, old_v;

  old_h = rootprops.harea_count;
  old_v = rootprops.varea_count;
  
  if (Get_Window_Property(RootWindow(dpy,screen), STPH_WIN_AREA_COUNT,
                          &length, &type, &size, &val)) {
          rootprops.harea_count = 1;
          rootprops.varea_count = 1;
  } else {
          rootprops.harea_count = ((long *) val)[0];
          rootprops.varea_count = ((long *) val)[1];
          XFree(val);
  }

  if (callbacks[STPH_AREA_COUNT_CB].func && !inhibit_callbacks
      && (old_h != rootprops.harea_count || old_v != rootprops.varea_count))
          RUN_CALLBACK(STPH_AREA_COUNT_CB,&rootprops);
}

static void Update_Win_Workspace(int inhibit_callbacks)
{
  char *val;
  long length;
  Atom type;
  int size;

  int old;

  old = rootprops.workspace;
  
  if (Get_Window_Property(RootWindow(dpy,screen), STPH_WIN_WORKSPACE,
                          &length, &type, &size, &val)) {
          rootprops.workspace = 0;
  } else {
          rootprops.workspace = (long) * (long *) val;
          XFree(val);
  }

  if (callbacks[STPH_WORKSPACE_CB].func && !inhibit_callbacks && old != rootprops.workspace)
          RUN_CALLBACK(STPH_WORKSPACE_CB,&rootprops);
}

static void Update_Win_Workspace_Count(int inhibit_callbacks)
{
  char *val;
  long length;
  Atom type;
  int size;

  int old;

  old = rootprops.workspace_count;
  
  if (Get_Window_Property(RootWindow(dpy,screen), STPH_WIN_WORKSPACE_COUNT,
                          &length, &type, &size, &val)) {
          rootprops.workspace_count = 1;
  } else {
          rootprops.workspace_count = (long) * (long *) val;
          XFree(val);
  }
  
  if (callbacks[STPH_WORKSPACE_COUNT_CB].func && !inhibit_callbacks
      && old != rootprops.workspace_count)
          RUN_CALLBACK(STPH_WORKSPACE_COUNT_CB,&rootprops);
}

void stph_connect(int cb_type, void (*func)(), void *data)
{
  callbacks[cb_type].func = func;
  callbacks[cb_type].data = data;
}

static int check_errors(Display *disp, XErrorEvent *xerr) /* this sucks */
{
  last_error.flag = 1;
  last_error.err = *xerr;
  return 0;
}

static void print_last_error(void)
{
  static char buf[100];
  last_error.flag = 0;
  fprintf(stderr, "***** stph: X error went unnoticed:\n");
  XGetErrorText(dpy, last_error.err.error_code, buf, 100);
  fprintf(stderr, "      %s\n", buf);
  fprintf(stderr, "      Please report this if you can reproduce it!\n");
}

int stph_initialise(void)
{
  dpy = XOpenDisplay(NULL); /* Posix-compliant systems should behave sensibly with NULL here; */
  if (dpy == NULL)          /* but probably we should try harder for those that don't... */
          return 1;
  screen = DefaultScreen(dpy);

  Setup_Atoms();

  XSetErrorHandler(check_errors);
  
  Update_Compliant_WM();
  if (!rootprops.compliant_wm)
          return 2;
  
  Update_Supported_Protocols();
  if (!(rootprops.supported_protocols && STPH_WIN_CLIENT_LIST_PCOL))
          return 3; /* can't deal with not having this yet; FIXME: add */
                    /* support for walking the tree manually */

  /* we have to suppress callbacks and call them manually here, because they aren't called if the
     old value is the same as the new -- but the old value is undefined at this stage */
  Update_Win_Workspace_Count(True);
  if (callbacks[STPH_WORKSPACE_COUNT_CB].func)
          RUN_CALLBACK(STPH_WORKSPACE_COUNT_CB, &rootprops);

  Update_Win_Workspace(True);
  if (callbacks[STPH_WORKSPACE_CB].func)
          RUN_CALLBACK(STPH_WORKSPACE_CB, &rootprops);

  Update_Win_Area_Count(True);
  if (callbacks[STPH_AREA_COUNT_CB].func)
          RUN_CALLBACK(STPH_AREA_COUNT_CB, &rootprops);

  Update_Win_Area(True);
  if (callbacks[STPH_AREA_CB].func)
          RUN_CALLBACK(STPH_AREA_CB, &rootprops);

  
  tasks_context = XUniqueContext();
  
  numtasks = 0;
  if(Refresh_Tasks())
          return 3;
  
  XSelectInput(dpy, RootWindow(dpy,screen), PropertyChangeMask | StructureNotifyMask);
  XSync(dpy, False);
  
  
  
  return 0;
}


static int Process_Event(XEvent *ev)
{
  
  switch(ev->type) {
  case PropertyNotify:
          /* alas we can't use a switch because the atoms don't all have their values set at compile */
          /* time */
          if (ev->xproperty.atom == XA_WM_NAME) {
                  Update_Task_Name(stph_task_from_xwin(ev->xproperty.window), False);
          }
          else if (ev->xproperty.atom == XA_WM_ICON_NAME) {
                  Update_Task_Icon_Name(stph_task_from_xwin(ev->xproperty.window), False);
          }
          else if (ev->xproperty.atom == XA_WM_CLASS) {
                  Update_Task_Class(stph_task_from_xwin(ev->xproperty.window), False);
          }
          else if (ev->xproperty.atom == XA_WM_HINTS) {
                  Update_Task_WM_Hints(stph_task_from_xwin(ev->xproperty.window), False);
          }
          else if (ev->xproperty.atom == STPH_WIN_SUPPORTING_WM_CHECK) {
                  Update_Compliant_WM();
          }
          else if (ev->xproperty.atom == STPH_WIN_PROTOCOLS) {
                  Update_Supported_Protocols();
          }
          else if (ev->xproperty.atom == STPH_WIN_LAYER) {
                  Update_Task_Layer(stph_task_from_xwin(ev->xproperty.window), False);
          }
          else if (ev->xproperty.atom == STPH_WIN_STATE) {
                  Update_Task_State(stph_task_from_xwin(ev->xproperty.window), False);
          }
          else if (ev->xproperty.atom == STPH_WIN_HINTS) {
                  Update_Task_Hints(stph_task_from_xwin(ev->xproperty.window), False);
          }
          else if (ev->xproperty.atom == STPH_WIN_APP_STATE) {
                  /* FIXME: gwmh uses this, but it's not mentioned in the Gnome WM Compliance spec
                     -- what is it?  Anyone need it?  */
          }
          else if (ev->xproperty.atom == STPH_WIN_EXPANDED_SIZE) {
                  /* FIXME: not yet supported; what is this?  anyone need it? */
          }
          else if (ev->xproperty.atom == STPH_KWM_WIN_ICON) {
                  Update_Task_KWM_Icon(stph_task_from_xwin(ev->xproperty.window), False);
          }
          else if (ev->xproperty.atom == STPH_WIN_ICONS) {
                  /* FIXME: not yet supported (not supported by Sawmill anyhow) */
          }
          else if (ev->xproperty.atom == STPH_WIN_WORKSPACE) {
                  if (ev->xproperty.window == RootWindow(dpy, screen)) {
                          Update_Win_Workspace(False);
                  } else {
                          Update_Task_Desktop(stph_task_from_xwin(ev->xproperty.window), False);
                  }
          }
          else if (ev->xproperty.atom == STPH_WIN_WORKSPACE_COUNT) {
                  Update_Win_Workspace_Count(False);
          }
          else if (ev->xproperty.atom == STPH_WIN_WORKSPACE_NAMES) {
                  /* FIXME: not yet supported */
          }
          else if (ev->xproperty.atom == STPH_WIN_CLIENT_LIST) {
                  Refresh_Tasks();
          }
          else if (ev->xproperty.atom == STPH_WIN_AREA) {
                  if (ev->xproperty.window == RootWindow(dpy, screen)) {
                          Update_Win_Area(False);
                  } else {
                          Update_Task_Area(stph_task_from_xwin(ev->xproperty.window), False);
                  }
          }
          else if (ev->xproperty.atom == STPH_WIN_AREA_COUNT) {
                  Update_Win_Area_Count(False);
          }
          else if (ev->xproperty.atom == STPH_WIN_UNIFIED_AREA) {
                  /* FIXME: what does this represent?  it's not in the gnome wm spec... */
          }
          else if (ev->xproperty.atom == XA_WM_STATE) {
                  Update_Task_Iconified(stph_task_from_xwin(ev->xproperty.window), False);
          }
          return 0;
  case FocusIn:
          Update_Task_Focused(stph_task_from_xwin(ev->xany.window), False);
          return 0;
  case FocusOut:
          Update_Task_Focused(stph_task_from_xwin(ev->xany.window), False);
          return 0;
  case CirculateNotify:
          /* we don't care about this... do we? */
          return 0;
  case ConfigureNotify:
          /* better off not using the results that come back with these xconfigure events */
          Update_Task_Win(stph_task_from_xwin(ev->xany.window), False);
          Update_Task_Frame(stph_task_from_xwin(ev->xany.window), False);
          
/*           printf("ConfigureNotify received: %ld, x:%d, y:%d, width:%d, height:%d\n", */
/*                  ev->xany.window, ev->xconfigure.x, ev->xconfigure.y, ev->xconfigure.width, */
/*                  ev->xconfigure.height); */
/*           printf("   above:%ld, override_redirect:%d\n", ev->xconfigure.above, */
/*                  ev->xconfigure.override_redirect); */
          return 0;
  case DestroyNotify:
          /* we don't care about this... do we? */
          return 0;
  case GravityNotify:
          /* we don't care about this... do we? */
          return 0;
  case MapNotify:
          /* Rely on _WIN_CLIENT_LIST for now... */
          return 0;
  case UnmapNotify:
          /* Rely on _WIN_CLIENT_LIST for now... */
          return 0;
  case ReparentNotify:
          /* we don't care about this... do we? */
          return 0;
  }
  return 0;
}

static int wm_protocol_supported_p(Window xwin, Atom a)
{
  Atom *protocols;
  int count, i;

  if (!(XGetWMProtocols(dpy, xwin, &protocols, &count))) {
          if (last_error.flag)
                  last_error.flag = 0; /* ignore BadWindow if generated */
          return 0;
  }
  
  for (i=0; i<count; i++) {
          if (protocols[i] == a) {
                  XFree(protocols);
                  return 1;
          }
  }
  XFree(protocols);
  return 0;
}

/* unashamedly ripped and adapted from gwmh.c, thanks Tim :) */
static int send_client_message_32 (Window recipient,
                                   Window event_window,
                                   Atom message_type,
                                   long event_mask,
                                   int n_longs,
                                   ...)
{
  XEvent xevent = { 0, };
  int i = 0;
  va_list var_args;

  if (n_longs > 5)
          return 1;
  
  va_start (var_args, n_longs);

  xevent.type = ClientMessage;
  xevent.xclient.window = event_window;
  xevent.xclient.message_type = message_type;
  xevent.xclient.format = 32;
  while (n_longs--)
    xevent.xclient.data.l[i++] = va_arg (var_args, long);

  va_end (var_args);

  if (!(XSendEvent (dpy, recipient, False, event_mask, &xevent)))
          return 1;
  if (last_error.flag) {
          last_error.flag = 0;
          return 1;
  }
  
  return 0;
}

int stph_set_current_desktop(long desktop)
{
  return send_client_message_32(RootWindow(dpy,screen),
                                RootWindow(dpy,screen),
                                STPH_WIN_WORKSPACE,
                                SubstructureNotifyMask,
                                2,
                                desktop,
                                CurrentTime);
}

int stph_set_current_area(long harea, long varea)
{
  return send_client_message_32(RootWindow(dpy,screen),
                                RootWindow(dpy,screen),
                                STPH_WIN_AREA,
                                SubstructureNotifyMask,
                                3,
                                harea,
                                varea,
                                CurrentTime);
}

int stph_unshade_task(struct StphTask *t)
{
  return send_client_message_32(RootWindow(dpy,screen),
                                t->xwin,
                                STPH_WIN_STATE,
                                SubstructureNotifyMask,
                                3,
                                STPH_STATE_SHADED,
                                0,
                                CurrentTime);
}

int stph_shade_task(struct StphTask *t)
{
  return send_client_message_32(RootWindow(dpy,screen),
                                t->xwin,
                                STPH_WIN_STATE,
                                SubstructureNotifyMask,
                                3,
                                STPH_STATE_SHADED,
                                STPH_STATE_SHADED,
                                CurrentTime);
}

int stph_shade_task_toggle(struct StphTask *t)
{
  Update_Task_State(t, True);
  
  if (t->state & STPH_STATE_SHADED)
          return stph_unshade_task(t);
  else
          return stph_shade_task(t);
}

int stph_raise_task_window(struct StphTask *t)
{
  XRaiseWindow(dpy, t->xwin);
  if (last_error.flag) {
          last_error.flag = 0;
          return 1;
  }
  return 0;
}

int stph_focus_task(struct StphTask *t)
{
  /* we need to make sure the task isn't shaded first */
  if (t->state & STPH_STATE_SHADED)
          if (stph_unshade_task(t))
                  return 1;
  
  /* we need to switch to the correct desktop first */
  if (t->desktop != rootprops.workspace)
          if (stph_set_current_desktop(t->desktop))
                  return 1;
  
  /* also, switch to the correct viewport */
  if (t->harea != rootprops.harea || t->varea != rootprops.varea)
          if (stph_set_current_area(t->harea, t->varea))
                  return 1;
  
  /* window must be mapped before it can be given focus */
  if (stph_uniconify_task(t))
          return 1;
  
  /* okay, now try to focus task */
  if (wm_protocol_supported_p(t->xwin, XA_WM_TAKE_FOCUS))
          send_client_message_32(t->xwin,
                                 t->xwin,
                                 0,
                                 2,
                                 XA_WM_TAKE_FOCUS,
                                 CurrentTime);
  Update_Task_Focused(t, True);
  /* and if at first you don't succeed... */
  if (!(t->focused)) {
          XSetInputFocus(dpy, t->xwin, RevertToParent, CurrentTime);
          if (last_error.flag) {
                  last_error.flag = 0;
                  return 1;
          }
  }
  return 0;
}

int stph_iconify_task(struct StphTask *t)
{
  if (!(XIconifyWindow(dpy, t->xwin, screen)))
          return 1;
  return 0;
}

int stph_uniconify_task(struct StphTask *t)
{
  XMapWindow(dpy, t->xwin);
  if (last_error.flag) {
          last_error.flag = 0;
          return 1;
  }
  return 0;
}

int stph_iconify_task_toggle(struct StphTask *t)
{
  Update_Task_Iconified(t, True);
  
  if (t->iconified)
          return stph_uniconify_task(t);
  else
          return stph_iconify_task(t);
}

int stph_close_task(struct StphTask *t)
{
  if (wm_protocol_supported_p(t->xwin, XA_WM_DELETE_WINDOW))
          return send_client_message_32(t->xwin,
                                        t->xwin,
                                        0,
                                        2,
                                        XA_WM_DELETE_WINDOW,
                                        CurrentTime);
  return 1;
}

int stph_kill_task(struct StphTask *t)
{
  XKillClient(dpy, t->xwin);
  if (last_error.flag) {
          last_error.flag = 0;
          return 1;
  }
  return 0;
}

int stph_close_or_kill_task(struct StphTask *t)
{
  if (!(stph_close_task(t)))
          return stph_kill_task(t);
  return 0;
}

int stph_move_task(struct StphTask *t, int x, int y)
{
  /* this takes into account the frame offset so that eg.
     the frame does not go off-screen if you move to (0,0) */
  XMoveWindow(dpy, t->xwin, x + t->win_x, y + t->win_y);
  if (last_error.flag) {
          last_error.flag = 0;
          return 1;
  }
  return 0;
}

void stph_mainiteration(void)
{
  XEvent ev;
  
  while (XPending(dpy)) {
          XNextEvent(dpy, &ev);
          Process_Event(&ev);
          if (last_error.flag) {
                  print_last_error();
          }
  }
}

