/* stphmodule.c -- Sawmill-inspired Taskbar and Pager Helper

   Copyright (C) 2000 Rob Hodges <s323140@student.uq.edu.au>

   This file is part of stph.

   stph is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   stph is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with stph; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <Python.h>
#include "stph.h"
#include "stph_icons.h"

/* ---------- structs ---------- */
typedef struct {
        PyObject *func;
        PyObject *data;
} pyfunc_and_data;

/* ---------- globals ---------- */

static pyfunc_and_data pycallbacks[STPH_MAX_ROOT_CALLBACK+1];


/* ---------- root objects ---------- */

/* root object struct */

typedef struct {
        PyObject_HEAD
        struct StphRoot *r;
} rootobject;

staticforward PyTypeObject Roottype;

/* root methods */
/* there are none */

/* root method registration table */
/* no methods; no registration table */

/* basic type-operations for root objects */

static rootobject *newrootobject()
{
  rootobject *self;
  self = PyObject_NEW(rootobject, &Roottype);
  if (self == NULL)
          return NULL;
  return self;
}

static void root_dealloc(rootobject *self)
{
  PyMem_DEL(self);
}

static int root_print(rootobject *self, FILE *fp, int flags)
{
  fprintf(fp, "[Root object]");
  return 0;
}

static PyObject *root_getattr(rootobject *self, char *attr)
{
  if (!strcmp(attr, "area") || !strcmp(attr, "viewport"))
          return Py_BuildValue("(i,i)", self->r->harea, self->r->varea);
  if (!strcmp(attr, "area_count") || !strcmp(attr, "viewport_count"))
          return Py_BuildValue("(i,i)", self->r->harea_count, self->r->varea_count);
  if (!strcmp(attr, "desktop") || !strcmp(attr, "workspace"))
          return Py_BuildValue("i", self->r->workspace);
  if (!strcmp(attr, "desktop_count") || !strcmp(attr, "workspace_count"))
          return Py_BuildValue("i", self->r->workspace_count);
  else
          return NULL;
  /* else */
/*           return Py_FindMethod(root_methods, (PyObject *)self, attr); */
}

/* root type descriptor */

static PyTypeObject Roottype = {
        PyObject_HEAD_INIT(&PyType_Type)
        0, /* ob_size */
        "root", /* tp_name */
        sizeof(rootobject), /* tp_basicsize */
        0, /* tp_itemsize */

        (destructor) root_dealloc,
        (printfunc) root_print,
        (getattrfunc) root_getattr,
        (setattrfunc) 0,
        (cmpfunc) 0,
        (reprfunc) 0,
        /* FIXME?  more go here? */
};



/* ---------- task objects ---------- */

/* task object struct */

typedef struct {
        PyObject_HEAD
        struct StphTask *t;
} taskobject;

staticforward PyTypeObject Tasktype;


/* task methods */

static PyObject *task_focus(taskobject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ":task.focus"))
          return NULL;
  return Py_BuildValue("i", stph_focus_task(self->t));
}

static PyObject *task_raise_window(taskobject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ":task.raise_window"))
          return NULL;
  return Py_BuildValue("i", stph_raise_task_window(self->t));
}

static PyObject *task_iconify(taskobject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ":task.iconify"))
          return NULL;
  return Py_BuildValue("i", stph_iconify_task(self->t));
}

static PyObject *task_uniconify(taskobject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ":task.uniconify"))
          return NULL;
  return Py_BuildValue("i", stph_uniconify_task(self->t));
}

static PyObject *task_iconify_toggle(taskobject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ":task.iconify_toggle"))
          return NULL;
  return Py_BuildValue("i", stph_iconify_task_toggle(self->t));
}

static PyObject *task_close(taskobject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ":task.close"))
          return NULL;
  return Py_BuildValue("i", stph_close_task(self->t));
}

static PyObject *task_kill(taskobject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ":task.kill"))
          return NULL;
  return Py_BuildValue("i", stph_kill_task(self->t));
}

static PyObject *task_close_or_kill(taskobject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ":task.close_or_kill"))
          return NULL;
  return Py_BuildValue("i", stph_close_or_kill_task(self->t));
}

static PyObject *task_move(taskobject *self, PyObject *args)
{
  int x,y;
  
  if (!PyArg_ParseTuple(args, "ii:task.move", &x, &y))
          return NULL;
  return Py_BuildValue("i", stph_move_task(self->t, x, y));
}

static PyObject *task_shade(taskobject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ":task.shade"))
          return NULL;
  return Py_BuildValue("i", stph_shade_task(self->t));
}

static PyObject *task_unshade(taskobject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ":task.unshade"))
          return NULL;
  return Py_BuildValue("i", stph_unshade_task(self->t));
}

static PyObject *task_shade_toggle(taskobject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ":task.shade_toggle"))
          return NULL;
  return Py_BuildValue("i", stph_shade_task_toggle(self->t));
}

/* task method registration table */

static struct PyMethodDef task_methods[] = {
        {"focus", (void *)task_focus, 1},
        {"raise_window", (void *)task_raise_window, 1},
        {"iconify", (void *)task_iconify, 1},
        {"uniconify", (void *)task_uniconify, 1},
        {"iconify_toggle", (void *)task_iconify_toggle, 1},
        {"close", (void *)task_close, 1},
        {"kill", (void *)task_kill, 1},
        {"close_or_kill", (void *)task_close_or_kill, 1},
        {"move", (void *)task_move, 1},
        {"shade", (void *)task_shade, 1},
        {"unshade", (void *)task_unshade, 1},
        {"shade_toggle", (void *)task_shade_toggle, 1},
        {NULL, NULL}
};

/* basic type-operations for tasks */

static taskobject *newtaskobject()
{
  taskobject *self;
  self = PyObject_NEW(taskobject, &Tasktype);
  if (self == NULL)
          return NULL;
  return self;
}

static void task_dealloc(taskobject *self)
{
  PyMem_DEL(self);
}

static int task_print(taskobject *self, FILE *fp, int flags)
{
  fprintf(fp, "[Task: %s]", self->t->name);
  return 0;
}

static PyObject *task_getattr(taskobject *self, char *attr)
{
  if (!strcmp(attr, "name"))
          return Py_BuildValue("s", self->t->name);
  if (!strcmp(attr, "icon_name"))
          return Py_BuildValue("s", self->t->icon_name);
  if (!strcmp(attr, "wm_class"))
          return Py_BuildValue("s", self->t->class);
  if (!strcmp(attr, "instance"))
          return Py_BuildValue("s", self->t->instance);
  if (!strcmp(attr, "focused"))
          return Py_BuildValue("i", self->t->focused);
  if (!strcmp(attr, "iconified"))
          return Py_BuildValue("i", self->t->iconified);
  if (!strcmp(attr, "frame_pos"))
          return Py_BuildValue("(i,i)", self->t->frame_x, self->t->frame_y);
  if (!strcmp(attr, "frame_dims"))
          return Py_BuildValue("(i,i)", self->t->frame_width, self->t->frame_height);
  if (!strcmp(attr, "win_pos"))
          return Py_BuildValue("(i,i)", self->t->win_x, self->t->win_y);
  if (!strcmp(attr, "win_dims"))
          return Py_BuildValue("(i,i)", self->t->win_width, self->t->win_height);
  if (!strcmp(attr, "desktop") || !strcmp(attr, "workspace"))
          return Py_BuildValue("i", self->t->desktop);
  if (!strcmp(attr, "area") || !strcmp(attr, "viewport"))
          return Py_BuildValue("(i,i)", self->t->harea, self->t->varea);
  if (!strcmp(attr, "layer"))
          return Py_BuildValue("i", self->t->layer);
  if (!strcmp(attr, "xwin"))
          return Py_BuildValue("i", self->t->xwin);
  if (!strcmp(attr, "xframe"))
          return Py_BuildValue("i", self->t->xframe);
  if (!strcmp(attr, "icon_pixmap"))
          return Py_BuildValue("i", self->t->icon_pixmap);
  if (!strcmp(attr, "icon_mask"))
          return Py_BuildValue("i", self->t->icon_mask);
  if (!strcmp(attr, "icon_window"))
          return Py_BuildValue("i", self->t->icon_window);
  if (!strcmp(attr, "kwm_icon_pixmap"))
          return Py_BuildValue("i", self->t->kwm_icon_pixmap);
  if (!strcmp(attr, "kwm_icon_mask"))
          return Py_BuildValue("i", self->t->kwm_icon_mask);
  /* state attributes */
  if (!strcmp(attr, "sticky"))
          return Py_BuildValue("i", self->t->state & STPH_STATE_STICKY);
  if (!strcmp(attr, "minimized") || !strcmp(attr, "minimised"))
          return Py_BuildValue("i", self->t->state & STPH_STATE_MINIMIZED);
  if (!strcmp(attr, "maximized_vert") || !strcmp(attr, "maximised_vert"))
          return Py_BuildValue("i", self->t->state & STPH_STATE_MAXIMIZED_VERT);
  if (!strcmp(attr, "maximized_horiz") || !strcmp(attr, "maximised_horiz"))
          return Py_BuildValue("i", self->t->state & STPH_STATE_MAXIMIZED_HORIZ);
  if (!strcmp(attr, "hidden"))
          return Py_BuildValue("i", self->t->state & STPH_STATE_HIDDEN);
  if (!strcmp(attr, "shaded"))
          return Py_BuildValue("i", self->t->state & STPH_STATE_SHADED);
  if (!strcmp(attr, "hid_workspace"))
          return Py_BuildValue("i", self->t->state & STPH_STATE_HID_WORKSPACE);
  if (!strcmp(attr, "hid_transient"))
          return Py_BuildValue("i", self->t->state & STPH_STATE_HID_TRANSIENT);
  if (!strcmp(attr, "fixed_position"))
          return Py_BuildValue("i", self->t->state & STPH_STATE_FIXED_POSITION);
  if (!strcmp(attr, "arrange_ignore"))
          return Py_BuildValue("i", self->t->state & STPH_STATE_ARRANGE_IGNORE);
  /* hint attributes */
  if (!strcmp(attr, "skip_focus"))
          return Py_BuildValue("i", self->t->hints & STPH_HINTS_SKIP_FOCUS);
  if (!strcmp(attr, "skip_winlist"))
          return Py_BuildValue("i", self->t->hints & STPH_HINTS_SKIP_WINLIST);
  if (!strcmp(attr, "skip_taskbar"))
          return Py_BuildValue("i", self->t->hints & STPH_HINTS_SKIP_TASKBAR);
  if (!strcmp(attr, "group_transient"))
          return Py_BuildValue("i", self->t->hints & STPH_HINTS_GROUP_TRANSIENT);
  if (!strcmp(attr, "focus_on_click"))
          return Py_BuildValue("i", self->t->hints & STPH_HINTS_FOCUS_ON_CLICK);
  if (!strcmp(attr, "do_not_cover"))
          return Py_BuildValue("i", self->t->hints & STPH_HINTS_DO_NOT_COVER);
  else
          return Py_FindMethod(task_methods, (PyObject *)self, attr);
}

static int task_compare(taskobject *v, taskobject *w)
{
  if (v->t->xwin < w->t->xwin)
          return -1;
  if (v->t->xwin > w->t->xwin)
          return 1;
  else
          return 0;
}

/* task type descriptor */

static PyTypeObject Tasktype = {
        PyObject_HEAD_INIT(&PyType_Type)
        0, /* ob_size */
        "task", /* tp_name */
        sizeof(taskobject), /* tp_basicsize */
        0, /* tp_itemsize */

        (destructor) task_dealloc,
        (printfunc) task_print,
        (getattrfunc) task_getattr,
        (setattrfunc) 0,
        (cmpfunc) task_compare,
        (reprfunc) 0,
        /* FIXME?  more go here? */
};


/* ---------- main module stuff ---------- */

/* module methods */

static PyObject *roottype_new(PyObject *self, PyObject *args)
{
  rootobject *ret;
  
  if (!PyArg_ParseTuple(args, ":root"))
          return NULL;
  ret = newrootobject();
  ret->r = stph_get_root_props();
  return (PyObject *)ret;
}


static PyObject *tasktype_new(PyObject *self, PyObject *args)
{
  Window xwin;
  taskobject *ret;
  
  if (!PyArg_ParseTuple(args, "l", &xwin))
          return NULL;
  
  ret = newtaskobject();
  if ((ret->t = stph_task_from_xwin(xwin)) == NULL) {
          fprintf(stderr, "stphmodule: no such window: %ld\n", xwin);
          return NULL; /* no such window */ 
  }
  return (PyObject *)ret;
}


static PyObject *wrap_initialise(PyObject *self, PyObject *args)
{
  int retval;

  if (!PyArg_ParseTuple(args, ":stph_initialise"))
          return NULL;
  retval = stph_initialise();
  return Py_BuildValue("i", retval);
  /* FIXME: maybe we should raise exceptions instead? */
}

static PyObject *wrap_mainiteration(PyObject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ":stph_mainiteration"))
          return NULL;
  
  stph_mainiteration();
  
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *wrap_set_current_desktop(PyObject *self, PyObject *args)
{
  long desktop;
  
  if (!PyArg_ParseTuple(args, "l:stph_set_current_desktop", &desktop))
          return NULL;
  return Py_BuildValue("i", stph_set_current_desktop(desktop));
}

static PyObject *wrap_set_current_area(PyObject *self, PyObject *args)
{
  long harea, varea;
  
  if (!PyArg_ParseTuple(args, "ll:stph_set_current_desktop", &harea, &varea))
          return NULL;
  return Py_BuildValue("i", stph_set_current_area(harea, varea));
}

static PyObject *wrap_xpm_data_from_drawable(PyObject *self, PyObject *args)
{
  Drawable win;
  Pixmap mask = 0;
  int width = -1, height = -1, nstrings, i;
  char **xpm_data;
  PyObject *ret;
  
  if (!PyArg_ParseTuple(args, "l|lll:stph_xpm_data_from_drawable",
                        &win, &mask, &width, &height))
          return NULL;
  xpm_data = stph_xpm_data_from_drawable(win, mask, width, height);
  if (xpm_data == NULL)
          return NULL; /* FIXME: this should probably raise an exception */

  nstrings = stph_get_xpm_data_length(xpm_data);

  ret = PyList_New(nstrings);
  for (i=0; i<nstrings; i++)
          PyList_SetItem(ret, i, PyString_FromString(xpm_data[i]));
  stph_free_xpm_data(xpm_data);
  
  return ret;
}


static void callback_dispatcher(void *t_or_r, int cb_type)
{
  PyObject *result;
  PyObject *temp;
  PyObject *obj, *arglist;
  
  if (cb_type <= STPH_MAX_TASK_CALLBACK) {
          obj = (PyObject *)newtaskobject();
          ((taskobject *)obj)->t = (struct StphTask *)t_or_r;
  } else {
          obj = (PyObject *)newrootobject();
          ((rootobject *)obj)->r = stph_get_root_props();
  }
  
  temp = PyTuple_New(1);
  PyTuple_SetItem(temp, 0, obj);
  if (pycallbacks[cb_type].data) {
          arglist = PySequence_Concat(temp, pycallbacks[cb_type].data);
          Py_DECREF(temp);
  } else {
          arglist = temp;
  }
  
  if (arglist == NULL) {
          fprintf(stderr, "stphmodule: error: callback arguments are NULL\n");
          PyErr_Clear();
          return;
  }
  
  result = PyEval_CallObject(pycallbacks[cb_type].func, arglist);
  if (result == NULL) {
          fprintf(stderr, "stphmodule: error: unable to run callback %d\n", cb_type);
          if (PyErr_Occurred())
                  PyErr_Print();
          return;
  }

  Py_DECREF(result);
}


static PyObject *wrap_connect(PyObject *self, PyObject *args)
{
  PyObject *result = NULL;
  PyObject *func, *data = NULL;
  int cb_type;

  if (PyArg_ParseTuple(args, "iO|O!:set_callback", &cb_type, &func, &PyTuple_Type, &data)) {
          if (!PyCallable_Check(func)) {
                  PyErr_SetString(PyExc_TypeError, "parameter must be callable");
                  return NULL;
          }
          Py_INCREF(func); /* keep new callback allocated */
          Py_XDECREF(pycallbacks[cb_type].func); /* free old callback, if any */
          pycallbacks[cb_type].func = func;
          pycallbacks[cb_type].data = data;
          Py_XINCREF(data);
          stph_connect(cb_type, callback_dispatcher, (void *)cb_type);

          Py_INCREF(Py_None);
          result = Py_None;
  }
  return result;
}


/* module method registration table */

static PyMethodDef stphMethods[] = {
        {"initialise", wrap_initialise, METH_VARARGS},
        {"mainiteration", wrap_mainiteration, METH_VARARGS},
        {"connect", wrap_connect, METH_VARARGS},
        {"task", tasktype_new, METH_VARARGS},
        {"root", roottype_new, METH_VARARGS},
        {"set_current_desktop", wrap_set_current_desktop, METH_VARARGS},
        {"set_current_area", wrap_set_current_area, METH_VARARGS},
        {"xpm_data_from_drawable", wrap_xpm_data_from_drawable, METH_VARARGS},
        {NULL, NULL}
};

/* initialisation */

void initstph()
{
  PyObject *m, *d;
  
  m = Py_InitModule("stph", stphMethods);
  d = PyModule_GetDict(m);

  PyDict_SetItemString(d, "ADD_TASK_CB", PyInt_FromLong((long) STPH_ADD_TASK_CB));
  PyDict_SetItemString(d, "REMOVE_TASK_CB", PyInt_FromLong((long) STPH_REMOVE_TASK_CB));
  PyDict_SetItemString(d, "TASK_NAME_CB", PyInt_FromLong((long) STPH_TASK_NAME_CB));
  PyDict_SetItemString(d, "TASK_ICON_NAME_CB", PyInt_FromLong((long) STPH_TASK_ICON_NAME_CB));
  PyDict_SetItemString(d, "TASK_CLASS_CB", PyInt_FromLong((long) STPH_TASK_CLASS_CB));
  PyDict_SetItemString(d, "TASK_STATE_CB", PyInt_FromLong((long) STPH_TASK_STATE_CB));
  PyDict_SetItemString(d, "TASK_HINTS_CB", PyInt_FromLong((long) STPH_TASK_HINTS_CB));
  PyDict_SetItemString(d, "TASK_FOCUS_CB", PyInt_FromLong((long) STPH_TASK_FOCUS_CB));
  PyDict_SetItemString(d, "TASK_ICONIFY_CB", PyInt_FromLong((long) STPH_TASK_ICONIFY_CB));
  PyDict_SetItemString(d, "TASK_FRAME_CB", PyInt_FromLong((long) STPH_TASK_FRAME_CB));
  PyDict_SetItemString(d, "TASK_FRAME_DIMS_CB", PyInt_FromLong((long) STPH_TASK_FRAME_DIMS_CB));
  PyDict_SetItemString(d, "TASK_FRAME_POS_CB", PyInt_FromLong((long) STPH_TASK_FRAME_POS_CB));
  PyDict_SetItemString(d, "TASK_WIN_DIMS_CB", PyInt_FromLong((long) STPH_TASK_WIN_DIMS_CB));
  PyDict_SetItemString(d, "TASK_WIN_POS_CB", PyInt_FromLong((long) STPH_TASK_WIN_POS_CB));
  PyDict_SetItemString(d, "TASK_LAYER_CB", PyInt_FromLong((long) STPH_TASK_LAYER_CB));
  PyDict_SetItemString(d, "TASK_DESKTOP_CB", PyInt_FromLong((long) STPH_TASK_DESKTOP_CB));
  PyDict_SetItemString(d, "TASK_AREA_CB", PyInt_FromLong((long) STPH_TASK_AREA_CB));
  PyDict_SetItemString(d, "TASK_ICON_PIXMAP_CB", PyInt_FromLong((long) STPH_TASK_ICON_PIXMAP_CB));
  PyDict_SetItemString(d, "TASK_ICON_WINDOW_CB", PyInt_FromLong((long) STPH_TASK_ICON_WINDOW_CB));
  PyDict_SetItemString(d, "TASK_KWM_ICON_CB", PyInt_FromLong((long) STPH_TASK_KWM_ICON_CB));
  PyDict_SetItemString(d, "AREA_CB", PyInt_FromLong((long) STPH_AREA_CB));
  PyDict_SetItemString(d, "AREA_COUNT_CB", PyInt_FromLong((long) STPH_AREA_COUNT_CB));
  PyDict_SetItemString(d, "WORKSPACE_CB", PyInt_FromLong((long) STPH_WORKSPACE_CB));
  PyDict_SetItemString(d, "WORKSPACE_COUNT_CB", PyInt_FromLong((long) STPH_WORKSPACE_COUNT_CB));

  if (PyErr_Occurred())
          Py_FatalError("Can't initialise module stph :(");
}
