/* stph_icons.c -- Obtain icons as xpm data, resizing if desired.

   Copyright (C) 2000 Rob Hodges <s323140@student.uq.edu.au>

   This file is part of stph.

   stph is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   stph is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with stph; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */


#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include "stph_icons.h"

#ifndef MIN
#define MIN(a,b)  ((a) < (b) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a,b)  ((a) > (b) ? (a) : (b))
#endif

#ifndef M_PI
#define M_PI  3.14159265358979323846
#endif

/* ====================================================================== */
/* Typedefs... */
/* ====================================================================== */

/* singly-linked list node, for use in hash tables */
typedef struct slnode
{
        unsigned long key;
        void *data;
soetnuh saoetnuh saontuh ansoeuth aosenuht saotnuh aonsetuh saoenuth aonsteuh         struct slnode *next;
} slnode;

/* simple hash table */
typedef struct
{
        slnode *table;
        int size;
        char ulmax_used;
} htable;

/* pixel representation */
typedef struct
{
        unsigned char r, g, b, a;
} Pixel;

/* image representation */
typedef struct
{
        int width;
        int height;
        Pixel *data;
} Image;



/* ====================================================================== */
/* Prototypes... */
/* ====================================================================== */

/* gcc -Wall doesn't like the prototype in stdio.h and complains; this prototype is just to shut it
   up so I can continue taking notice of warnings that are actually meaningful */
extern int asprintf (char **PTR, const char *TEMPLATE, ...);

static void freecodes(char** codes, int numcols);
static char **gencodes(int numcols);
static void del_htable(htable *ht);
static void *htable_get(htable *ht, unsigned long key);
static int htable_add(htable *ht, unsigned long key, void *data);
static htable *make_htable(int size);
static Image *image_from_drawable(Drawable win, Pixmap mask);
static void __handle_x_error(Display * d, XErrorEvent * ev);
static Image *resample_image(const Image *src, int width, int height);
static Image *copy_image(const Image *src);
static void free_image(Image *im);
static Image *malloc_image(int width, int height);
static double lanczos_filter(double x);
static double sinc_filter(double x);
static void apply_filter_1d(const Image *src, Image *dest, double (*filter)(double),
                            double filter_support, char mode);
static char **xpm_data_from_image(const Image *im);

/* ====================================================================== */
/* basic Image create/delete/copy functions */
/* ====================================================================== */
static Image *malloc_image(int width, int height)
{
  Image *ret;
  ret = (Image *) malloc(sizeof(Image));
  ret->data = (Pixel *) malloc(sizeof(Pixel) * width * height);
  ret->width = width;
  ret->height = height;
  if (ret->data == NULL){
          free(ret);
          return NULL;
  }
  return ret;
}

static void free_image(Image *im)
{
  free(im->data);
  free(im);
}

static Image *copy_image(const Image *src)
{
  Image *ret;
  long i;
  ret = malloc_image(src->width, src->height);
  for (i=0; i < src->width * src->height; i++)
          ret->data[i] = src->data[i];
  return ret;
}

/* ====================================================================== */
/* Image resampling functions */
/* ====================================================================== */

static void apply_filter_1d(const Image *src, Image *dest, double (*filter)(double),
                            double filter_support, char mode)
{
  double *weights;
  double stretch_factor, shrink_factor, support, density, centre, r, g, b, a;
  long desired_length, desired_breadth, orig_length, pin_incr, pin_span,
          y_or_x, x_or_y, offset, i, n, pout_incr;
  int start, end;
  char switch1, switch2;
  Pixel *p_in, *p_out;
  
  /* set up some stuff the generalise the algorithm for either vertical or horiz. filtering */
  if (mode == 'v') { /* filtering vertically */
          stretch_factor = (double) dest->height / src->height;
          desired_length = dest->height;
          desired_breadth = dest->width;
          orig_length = src->height;
          pin_incr = src->width;
          pin_span = 1;
          switch1 = 0;
          switch2 = 1;
          pout_incr = 1;
  } else { /* filtering horizontally */
          stretch_factor = (double) dest->width / src->width;
          desired_length = dest->width;
          desired_breadth = dest->height;
          orig_length = src->width;
          pin_incr = 1;
          pin_span = src->width;
          switch1 = 1;
          switch2 = 0;
          pout_incr = dest->width;
  }
  /* determine size of supports (influential pixels on each side) */
  shrink_factor = MAX(1.0/stretch_factor, 1.0);
  support = MAX(shrink_factor*filter_support, 0.5);
  if (support <= 0.5) {
          support = 0.5;
          shrink_factor = 1.0;
  }
  support += 0.000001; /* fp imprecision will not bother us */

  /* use filter function to generate weights assigned to influential pixels */
  weights = (double *) malloc(sizeof(double)*orig_length);
  for (y_or_x = 0; y_or_x < desired_length; y_or_x++) {
          density = 0;
          centre = (double) y_or_x / stretch_factor;
          start = (int) MAX(centre-support+0.5, 0);
          end = (int) MIN(centre+support+0.5, orig_length);
          for (n=0, i=start; i < end; i++, n++) {
                  weights[n] = filter( ((double)i - centre + 0.5) / shrink_factor ) / shrink_factor;
                  density += weights[n];
          }
          /* if possible and necessary, normalise weights: */
          if (density != 0.0 && density != 1.0) 
                  for (i=0; i<n; i++)
                          weights[i] /= density;

          /* use weighted sums from influential pixels to determine output pixel values */
          p_in = src->data + start*switch1 + start*switch2*src->width;
          p_out = dest->data + y_or_x*switch1 + y_or_x*switch2*dest->width;
          for (x_or_y = 0; x_or_y < desired_breadth; x_or_y++) {
                  r = g = b = a = 0.0;
                  offset = x_or_y * pin_span;
                  for (i=0; i<n; i++) {
                          r += weights[i] * (p_in + offset)->r;
                          g += weights[i] * (p_in + offset)->g;
                          b += weights[i] * (p_in + offset)->b;
                          a += weights[i] * (p_in + offset)->a;
                          offset += pin_incr;
                  }
                  p_out->r = (r < 0) ? 0 : (r > 255) ? 255 : (unsigned char)(r + 0.5);
                  p_out->g = (g < 0) ? 0 : (g > 255) ? 255 : (unsigned char)(g + 0.5);
                  p_out->b = (b < 0) ? 0 : (b > 255) ? 255 : (unsigned char)(b + 0.5);
                  p_out->a = (a < 0) ? 0 : (a > 255) ? 255 : (unsigned char)(a + 0.5);
                  p_out += pout_incr;
          }
  }
  free(weights);
}

static double sinc_filter(double x)
{
  x *= M_PI;
  if (x != 0.0)
          return sin(x)/x;
  return 1.0;
}

static double lanczos_filter(double x)
{
  if (x < 0)
          x = -x;
  if (x < 3.0)
          return sinc_filter(x) * sinc_filter(x/3.00000000001);
  return 0.0;
}

static Image *resample_image(const Image *src, int width, int height)
{
  Image *interm, *dest;
  double (*filterfunc)(double);
  double filter_support;
  
  if (src == NULL || src->width == 0 || src->height == 0)
          return NULL;
  if (src->width == width && src->height == height)
          return copy_image(src);
  
  dest = malloc_image(width, height);
  
  filterfunc = lanczos_filter;
  filter_support = 3.0;
  
  if (height >= src->height) {
          interm = malloc_image(width, src->height);
          apply_filter_1d(src, interm, filterfunc, filter_support, 'h');
          apply_filter_1d(interm, dest, filterfunc, filter_support, 'v');
  } else {
          interm = malloc_image(src->width, height);
          apply_filter_1d(src, interm, filterfunc, filter_support, 'v');
          apply_filter_1d(interm, dest, filterfunc, filter_support, 'h');
  }
  free_image(interm);
  return dest;
}

/* ====================================================================== */
/* Drawable -> Image... */
/* ====================================================================== */

/* The following code for grabbing the data from a drawable has been largely
   stolen and modified from code in Imlib. */

static char x_error = 0;

static void __handle_x_error(Display * d, XErrorEvent * ev)
{
  d = NULL;
  ev = NULL;
  x_error = 1;
}

static Image *image_from_drawable(Drawable win, Pixmap mask)
{
  unsigned char      r, g, b;
  unsigned long      pixel;
  int                i, xx, yy, winx, winy, width, height, depth, border, pidx;
  int                screen_depth, render_depth, dummy;
  XVisualInfo        xvi, *xvir;
  Visual             *visual;
  XImage             *ximage;
  XWindowAttributes  xatt;
  Colormap           cmap;
  char               is_pixmap;
  Display            *dpy;
  Window             root;
  XColor             ctab[256];
  XErrorHandler      erh;
  Image              *ret;

  
  dpy = XOpenDisplay(NULL);
  XGrabServer(dpy);
  erh = XSetErrorHandler((XErrorHandler) __handle_x_error);
  x_error = 0;
  /* try to get width, height & depth of drawable */
  XGetWindowAttributes(dpy, win, &xatt);
  XFlush(dpy);
  if (x_error) {
          /* then win must be a pixmap, not a window */
          x_error = 0;
          is_pixmap = 1;
          XGetGeometry(dpy, win, &root, &winx, &winy,
                       (unsigned int*) &xatt.width,
                       (unsigned int*) &xatt.height,
                       &border,
                       (unsigned int *) &xatt.depth);
          XFlush(dpy);
          if (x_error) {
                  /* oh, win wasn't a pixmap either? bug out then... */
                  XUngrabServer(dpy);
                  XFlush(dpy);
                  XSetErrorHandler((XErrorHandler) erh);
                  return NULL;
          }
  }
  width = xatt.width;
  height = xatt.height;
  depth = xatt.depth;

  ret = malloc_image(width, height);
  
  ximage = XGetImage(dpy, win, 0,0, width, height, AllPlanes, ZPixmap);
  XUngrabServer(dpy);
  XFlush(dpy);

  /* determine screen and render depth */
  screen_depth = DefaultDepth(dpy, DefaultScreen(dpy));
  render_depth = screen_depth;
  if (screen_depth == 16) {
          visual = DefaultVisual(dpy, DefaultScreen(dpy));
          xvi.visual = visual;
          xvi.visualid = XVisualIDFromVisual(visual);
          xvir = XGetVisualInfo(dpy, VisualIDMask, &xvi, &dummy);
          if (xvir) {
                  if (xvir->red_mask != 0xf800)
                          render_depth = 15;
                  XFree(xvir);
          }
  }

  /* finally, get the data */
  if (depth == 1) {
          ctab[0].red = 255;
          ctab[0].green = 255;
          ctab[0].blue = 255;
          ctab[1].red = 0;
          ctab[1].green = 0;
          ctab[1].blue = 0;
  } else if (depth <= 8) {
          XColor              cols[256];
          
          if (!is_pixmap) {
                  cmap = xatt.colormap;
                  if (cmap == None)
                          cmap = DefaultColormap(dpy, DefaultScreen(dpy));
          } else
                  cmap = DefaultColormap(dpy, DefaultScreen(dpy));

          for (i = 0; i < (1 << screen_depth); i++) {
                  cols[i].pixel = i;
                  cols[i].flags = DoRed | DoGreen | DoBlue;
          }
          XQueryColors(dpy, cmap, cols, 1 << screen_depth);
          for (i = 0; i < (1 << screen_depth); i++) {
                  ctab[i].red = cols[i].red >> 8;
                  ctab[i].green = cols[i].green >> 8;
                  ctab[i].blue = cols[i].blue >> 8;
                  ctab[i].pixel = cols[i].pixel;
          }
  }

  
  if (ret) {
          pidx = 0;
          switch (depth) {
          case 0:
          case 1:
          case 2:
          case 3:
          case 4:
          case 5:
          case 6:
          case 7:
          case 8:
                  for (yy = 0; yy < height; yy++) {
                          for (xx = 0; xx < width; xx++) {
                                  pixel = XGetPixel(ximage, xx, yy);
                                  r = ctab[pixel & 0xff].red;
                                  g = ctab[pixel & 0xff].green;
                                  b = ctab[pixel & 0xff].blue;
                                  ret->data[pidx].r = r;
                                  ret->data[pidx].g = g;
                                  ret->data[pidx].b = b;
                                  pidx++;
                          }
                  }
                  break;
          case 16:
          case 15:
                  if (render_depth == 16) {
                          for (yy = 0; yy < height; yy++) {
                                  for (xx = 0; xx < width; xx++) {
                                          pixel = XGetPixel(ximage, xx, yy);
                                          r = (pixel >> 8) & 0xf8;
                                          g = (pixel >> 3) & 0xfc;
                                          b = (pixel << 3) & 0xf8;
                                          ret->data[pidx].r = r;
                                          ret->data[pidx].g = g;
                                          ret->data[pidx].b = b;
                                          pidx++;
                                  }
                          }
                  } else {
                          for (yy = 0; yy < height; yy++) {
                                  for (xx = 0; xx < width; xx++) {
                                          pixel = XGetPixel(ximage, xx, yy);
                                          r = (pixel >> 7) & 0xf8;
                                          g = (pixel >> 3) & 0xf8;
                                          b = (pixel << 3) & 0xf8;
                                          ret->data[pidx].r = r;
                                          ret->data[pidx].g = g;
                                          ret->data[pidx].b = b;
                                          pidx++;
                                  }
                          }
                  }
                  break;
          case 24:
          case 32:
                  for (yy = 0; yy < height; yy++) {
                          for (xx = 0; xx < width; xx++) {
                                  pixel = XGetPixel(ximage, xx, yy);
                                  r = (pixel >> 16) & 0xff;
                                  g = (pixel >> 8) & 0xff;
                                  b = pixel & 0xff;
                                  ret->data[pidx].r = r;
                                  ret->data[pidx].g = g;
                                  ret->data[pidx].b = b;
                                  pidx++;
                          }
                  }
                  break;
          default:
                  for (yy = 0; yy < height; yy++) {
                          for (xx = 0; xx < width; xx++) {
                                  r = rand();
                                  g = rand();
                                  b = rand();
                                  ret->data[pidx].r = r;
                                  ret->data[pidx].g = g;
                                  ret->data[pidx].b = b;
                                  pidx++;
                          }
                  }
                  break;
          }
  }

  XDestroyImage(ximage);
  
  if (!ret)
          return NULL;

  pidx = 0;
  if (mask) {
          ximage = XGetImage(dpy, mask, 0,0, width, height, AllPlanes, ZPixmap);
          for (yy = 0; yy < height; yy++)
                  for (xx = 0; xx < width; xx++)
                          ret->data[pidx++].a = (XGetPixel(ximage, xx, yy)) ? 255 : 0;
          XDestroyImage(ximage);
  } else {
          for (yy = 0; yy < height; yy++)
                  for (xx = 0; xx < width; xx++)
                          ret->data[pidx++].a = 255;
  }
  
  
  XSetErrorHandler((XErrorHandler) erh);
  XCloseDisplay(dpy);
  return ret;
}


/* ====================================================================== */
/* Hash table functions, to help generate xpm data quickly... */
/* ====================================================================== */

/* create hash table of a given SIZE */
static htable *make_htable(int size)
{
  htable *ret;
  int i;

  ret = (htable *) malloc(sizeof(htable));
  ret->table = (slnode *) malloc(size*sizeof(slnode));
  for (i=0; i<size; i++) {
          ret->table[i].key = ULONG_MAX; /* an unlikely key; saves cycles */
          ret->table[i].data = NULL;
          ret->table[i].next = NULL;
  }
  ret->size = size;
  ret->ulmax_used = 0;
  return ret;
}

/* Add KEYed DATA into hashtable.  DATA may be NULL (and if not, it will NOT be freed for you!).
   Returns 0 if KEY is already there (in which case the existing data corresponding to that key is
   replaced with DATA); returns 1 if not. */
static int htable_add(htable *ht, unsigned long key, void *data)
{
  int idx, retval;
  slnode *this, *last;
  unsigned long oldkey;
  
  idx = key % ht->size;
  oldkey = ht->table[idx].key;

  if (key == ULONG_MAX) {
          if (ht->ulmax_used)
                  retval = 0;
          else
                  retval = 1;
          ht->ulmax_used = 1;
  }
  
  if ( !(ht->ulmax_used && oldkey == ULONG_MAX && ULONG_MAX%(ht->size) == idx && key != ULONG_MAX)
       && (oldkey == ULONG_MAX || oldkey == key) ) {
          ht->table[idx].key = key;
          ht->table[idx].data = data;
          if (oldkey == key && key != ULONG_MAX)
                  retval = 0;
          if (oldkey == ULONG_MAX && key != ULONG_MAX)
                  retval = 1;
          return retval;
  }
  
  this = ht->table[idx].next;
  last = &(ht->table[idx]);
  while (this) {
          if (this->key == key) {
                  this->data = data;
                  return 0;
          }
          last = this;
          this = this->next;
  };
  last->next = (slnode *) malloc(sizeof(slnode));
  last->next->key = key;
  last->next->data = data;
  last->next->next = NULL;
  return 1;
}


/* return data associated with KEY in hash table HT */
static void *htable_get(htable *ht, unsigned long key)
{
  int idx;
  slnode *this;

  if (key == ULONG_MAX && !(ht->ulmax_used))
          return NULL;
  
  idx = key % ht->size;
  this = &(ht->table[idx]);
  do {
          if (this->key == key)
                  return this->data;
          this = this->next;
  } while (this);
  return NULL;
}

/* free memory associated with hash table HT */
static void del_htable(htable *ht)
{
  int i;
  slnode *this, *next;

  for (i=0; i<ht->size; i++) {
          if ((next = ht->table[i].next)) {
                  do {
                          this = next;
                          next = this->next;
                          free(this);
                  } while (next);
          }
  }
  free(ht->table);
  free(ht);
}

/* ====================================================================== */
/* Support functions for xpm data generation... */
/* ====================================================================== */

static char **gencodes(int numcols)
{
  char **ret;
  int i,j, numchars = 1;
  char *work;

  /* [33,126] printable ascii, not including space (32) -- 94 chars */
  while (numcols > (int) pow(94, numchars))
          numchars++;
  
  ret = (char **) malloc(numcols*sizeof(char *));
  for (i=0; i<numcols; i++)
          ret[i] = (char *) malloc((numchars+1)*sizeof(char));

  work = (char *) malloc((numchars+1)*sizeof(char));
  
  for (i=0; i<numchars; i++)
          work[i] = (char) 33;
  work[numchars] = '\0';
  
  for (i=0; i<numcols; i++) {
          strcpy(ret[i], work);
          work[numchars-1] += 1;
          for (j=numchars-1; j>=0; j--) {
                  if (work[j] > 126) {
                          work[j] = 33;
                          work[j-1] += 1;
                  } else {
                          break;
                  }
          }
  }
  free(work);
  return ret;
}

/* free the strings created by gencodes() above */
static void freecodes(char** codes, int numcols)
{
  int i;
  
  for (i=0; i<numcols; i++)
          free(codes[i]);
  free(codes);
}

/* ====================================================================== */
/* Image -> xpm_data ... */
/* ====================================================================== */
static char **xpm_data_from_image(const Image *im)
{
  char **ret;
  char *buf;
  char *Nonestr;
  int i, x,y,size, numcols=0, pixno=0, chars_per_pixel;
  char **codes;
  unsigned long colour;
  unsigned long *col_data, *unique_cols;
  htable *ht;

  
  size = im->width * im->height;
  
  col_data = (unsigned long *) malloc(sizeof(unsigned long)*size); //freed
  unique_cols = (unsigned long *) malloc(sizeof(unsigned long)*size); //freed
  
  ht = make_htable(size);
  for (i=0; i<size; i++) {
          colour = im->data[i].r << 16 | im->data[i].g << 8 | im->data[i].b;
          col_data[i] = colour;
          if (htable_add(ht, colour, NULL))
                  unique_cols[numcols++] = colour;
  }
  del_htable(ht);

  codes = gencodes(numcols);
  ht = make_htable(numcols*2);
  for (i=0; i<numcols; i++)
          htable_add(ht, unique_cols[i], codes[i]);
  chars_per_pixel = strlen(codes[0]);

  ret = (char **) malloc((2+numcols+im->height)*sizeof(char *));
  /* (2+numcols+height) because "None" is not included in numcols */
  
  asprintf(&buf, "%d %d %d %d", im->width, im->height, numcols+1, chars_per_pixel);
  ret[0] = buf;
  Nonestr = (char *) malloc(sizeof(char)*(chars_per_pixel+1)); //freed
  for (i=0; i<chars_per_pixel; i++)
          strcpy(Nonestr+i, " ");
  asprintf(&buf, "%s\tc None", Nonestr);
  ret[1] = buf;
  for (i=0; i<numcols; i++) {
          asprintf(&buf, "%s\tc #%02x%02x%02x",
                   (char *)htable_get(ht, unique_cols[i]),
                   (unsigned char) (unique_cols[i] >> 16),
                   (unsigned char) (unique_cols[i] >> 8),
                   (unsigned char) (unique_cols[i]));
          ret[i+2] = buf;
  }
  
  pixno = 0;
  for (y=0; y < im->height; y++) {
          buf = (char *) malloc(sizeof(char)*(chars_per_pixel * im->width + 1));
          for (x=0; x < im->width; x++, pixno++) {
                  if(im->data[pixno].a > 127)
                          strcpy(buf+x*chars_per_pixel,
                                 (char *)htable_get(ht, col_data[pixno]));
                  else
                          strcpy(buf+x*chars_per_pixel, Nonestr);
          }
          ret[y+2+numcols] = buf;
  }
  
  free(Nonestr);
  free(col_data);
  free(unique_cols);
  freecodes(codes, numcols);
  //print_htable(ht, 0);
  
  del_htable(ht);
  return ret;
}

/* ====================================================================== */
/* Drawable -> xpm_data ... */
/* ====================================================================== */
char **stph_xpm_data_from_drawable(Drawable d, Drawable mask, int desired_width, int desired_height)
{
  Image *im, *resized = NULL;
  char **ret;

  if (desired_width == 0 || desired_height == 0)
          return NULL;

  if (desired_width < 0 || desired_height < 0) {
          im = image_from_drawable(d, mask);
          if (im) {
                  ret = xpm_data_from_image(im);
                  free_image(im);
                  return ret; 
          }
          return NULL;
  }

  im = image_from_drawable(d, mask);
  if (im == NULL)
          return NULL;
  resized = resample_image(im, desired_width, desired_height);
  if (resized == NULL) {
          free_image(im);
          return NULL;
  }
  ret = xpm_data_from_image(resized);
  free_image(im);
  free_image(resized);
  return ret;
}

/* ====================================================================== */
/* Functions to aid in dealing with xpm data */
/* ====================================================================== */

/* returns the number of strings in DATA, which must be in XPM format */
long stph_get_xpm_data_length(char** data)
{
  long length; /* number of strings in DATA must be determined from first string,
                which is in format "width, height, n_colours, chars_per_pixel" */
  char *tmp, *bak;

  tmp = strdup(data[0]); /* strsep screws with data so make copy */
  bak = tmp; /* strsep screws with pointer too */

  strsep(&tmp, " "); /* discard width number in first string */
  length = 1 /* header line */
          + strtol(strsep(&tmp, " "), NULL, 10) /* no. lines of img data */
          + strtol(strsep(&tmp, " "), NULL, 10); /* no. lines of color data */

  free(bak);
  return length;
}

/* free the xpm data malloc()'d by stph_xpm_data_from_drawable() */
void stph_free_xpm_data(char** data)
{
  long length, i;

  length = stph_get_xpm_data_length(data);
  for(i=0; i<length; i++)
          free(data[i]);
  free(data);
}


/* prints the xpm data on stdout -- mainly for debugging */

/* void print_xpm_data(char **data) */
/* { */
/*   long length, i; */
/*   length = stph_get_xpm_data_length(data); */
/*   for (i=0; i<length; i++) */
/*           printf("%s\n", data[i]); */
/* } */


/* ====================================================================== */
/* ====================================================================== */
/* ====================================================================== */

/* int main(int argc, char **argv) */
/* { */
/*   char **xpm_d; */
  
/*   xpm_d = stph_xpm_data_from_drawable(0x2000052, 0, -1, -1); */

/*   if (xpm_d) */
/*           print_xpm_data(xpm_d); */

/*   if (xpm_d) */
/*           stph_free_xpm_data(xpm_d); */
/*   return 0; */
  
/* } */
